"""Defines all functions within plot part of exergi package."""

from pandas import DataFrame
from plotly import express
import plotly.graph_objects as go


def lineplot(df_: DataFrame, x: str, y: list, width=1200, 
             height=500, title='No title'):
    """Create a lineplot of columns in specified dataframe using plotly.

    Arguments:
        df_         - Dataframe with data to plot. All data must be in the 
                      same dataframe    
        x           - String with name of column for x-axis, can't be index
        y           - List of string with name of columns for y-axis.
        width       - Width of plot in pixels.
        height      - Height of plot in pixels.        
        title       - Title for plot.
        
    Returns:
        - Plotly lineplot with legend.
        
    """    
    # Define position of legend
    legend_dict = dict(orientation='h', yanchor='bottom', y=1.0, 
                       xanchor='left', x=0)
    
    # Create figure
    fig = express.line(df_, x=x, y=y, title=title, width=width, height=height)
    fig.update_layout(legend=legend_dict)
    fig.show()



def histogram(df_: DataFrame, x: list, width=1000, 
              height=500, title='No title'):
    """Create a histogram of columns in specified dataframe using plotly.

    Arguments:
        df_         - Dataframe with data to plot. All data must be in the 
                      same dataframe    
        x           - List of string with name of columns.
        width       - Width of plot in pixels.
        height      - Height of plot in pixels.        
        title       - Title for plot.
        
    Returns:
        - Plotly histogram
        
    """ 
    fig = go.Figure()

    for col in x:
        fig.add_trace(go.Histogram(x=df_[col], name=col))

    # Overlay both histograms
    fig.update_layout(barmode='overlay', 
                      title=title,
                      width=width,
                      height=height)

    # Reduce opacity to see both histograms
    fig.update_traces(opacity=0.6)
    fig.show()


"""Defines all functions within the s3 (simple storage service) module."""

# Standard library imports
from io import BytesIO, StringIO
from os.path import dirname, splitext
from tempfile import TemporaryFile, NamedTemporaryFile
from pickle import dumps
from logging import getLogger
from time import time

# Related third party imports 
from boto3 import client, resource
from numpy import load, save, savez
# from sklearn.externals import joblib
from pandas import read_csv, read_excel, read_pickle, read_hdf, \
    ExcelWriter, HDFStore


def export_file(obj, bucket: str, s3key: str, **kwargs) -> None:
    """Export python object to specified location in AWS S3.

    Keyword Arguments:
        - obj       - Python object to be exported
        - bucket    - S3 export bucket,no trailing "/"
        - s3key     - S3 export key, no leading "/". String should end with the
                      desired file format. "public/.../example.csv".
                      Currently supported fileformats are
                        - .csv
                            - pandas.DataFrame()
                        - .xlsx
                            - pandas.DataFrame()
                        - .pkl
                            - pandas.DataFrame()
                            - sklearn (temporarly removed)
                        - .h5
                            - pandas.DataFrame()
                        - .npy
                            - numpy (arrays)
                        - .npz
        - **kwargs
    """
    # Initiate Logger
    logger = getLogger(__name__)  
    start = time()

    # Connect to S3
    s3resource = resource("s3")

    # Extra file name, file format and export object type
    _, file_fmt = splitext(s3key)
    obj_class = obj.__class__.__module__.split(".")[0]

    logger.info("""Load settings:\n\tobj_class=%s\n\t bucket=%s\n\ts3key=%s""",
                obj_class, bucket, s3key)
    logger.info('Exporting %s-data to S3:', file_fmt)

    # Comma separated files
    if (file_fmt == ".csv") & (obj_class == "pandas"):
        buffer = StringIO()
        obj.to_csv(buffer, index=False)
        s3resource.Object(bucket, s3key).put(Body=buffer.getvalue())

    # Excel files
    elif (file_fmt == ".xlsx") & (obj_class == "pandas"):
        with BytesIO() as output:
            with ExcelWriter(output, engine="xlsxwriter") as writer:
                obj.to_excel(writer)
            data = output.getvalue()
            s3resource.Object(bucket, s3key).put(Body=data)

    # Pickle files
    # elif (file_fmt == ".pkl") & (obj_class == "sklearn"):
    #     with TemporaryFile() as temp_file:
    #         joblib.dump(obj, temp_file)
    #         temp_file.seek(0)
    #         s3resource.Object(bucket, s3key).put(Body=temp_file.read())

    elif (file_fmt == ".pkl") & (obj_class == "pandas"):
        serialized_data = dumps(obj)
        s3resource.Object(bucket, s3key).put(Body=serialized_data)

    # HDF5 files
    elif (file_fmt == ".h5") & (obj_class == "pandas"):
        with NamedTemporaryFile(suffix=".h5") as temp_file:
            hdf = HDFStore(temp_file.name)
            hdf.put(value=obj, format='table', data_columns=True, **kwargs)
            hdf.close()
            temp_file.seek(0)
            s3resource.Object(bucket, s3key).put(Body=temp_file.read())

    # npy files
    elif (file_fmt == ".npy") & (obj_class == "numpy"):
        with TemporaryFile() as temp_file:
            save(temp_file, obj)
            temp_file.seek(0)
            s3resource.Object(bucket, s3key).put(Body=temp_file.read())

    # npz files- This exports each pandas column to a compressed numpy array
    # as a "key-value pair". See numpy documentation for more information
    elif (file_fmt == ".npz") & (obj_class == "pandas"):
        with TemporaryFile() as temp_file:
            savez(temp_file, **obj.to_dict())
            temp_file.seek(0)
            s3resource.Object(bucket, s3key).put(Body=temp_file.read())

    else:
        raise_str = f"""Failed! {file_fmt} export of {obj_class} objects not
            implemented."""
        logger.info('\tFailed! %s', raise_str)
        raise Exception()

    logger.info('\tOK! Exported data in %f [s]', time() - start)


def import_file(bucket: str, s3key: str, obj_class: str = "pandas", **kwargs):
    """Import DataFrame from specified bucket/s3key.

    Function imports data from specified bucket/s3 key combination based on 
    the ending file format of the passed s3key. 

    Arguments:
        - bucket    -   S3 export bucket, no trailing "/"
        - s3key     -   S3 export key, no leading "/" String should end with
                        the desired file format like "public/.../example.csv"
                        Currently supported fileformats is:
                            - .csv
                                - pandas.DataFrame()
                            - .xlsx
                                - pandas.DataFrame()
                            - .pkl
                                - pandas.DataFrame()
                                - sklearn.joblib (temporarly removed)
                            - .h5
                                - pandas.DataFrame()
                            - .npy
                                - numpy (array)
                            - .npz
                                - numpy-zip (arrays)
        - obj_class -   String explaining what object type file should be
                        loaded as (default = "pandas")
    Keyword Arguments:
        - **kwargs  -   Keyword arguments import function. Import function
                        varies for each file format:
                            - .csv  = pd.read_csv()
                            - .xlsx = pd.read_excel()
                            - .pkl  = pd.read_pickle()
                            - .h5   = pd.read_hdf()
    Returns:
        - obj       -   Imported object
    """
    # Initiate Logger
    logger = getLogger(__name__)
    start = time()

    # Extra file name, file format and export object type
    _, file_fmt = splitext(s3key)
    s3client = client("s3")

    logger.info("""Load settings:\n\tobj_class=%s\n\t
        bucket=%s\n\ts3key=%s""", obj_class, bucket, s3key)
    logger.info('Loading %s-data from S3:', file_fmt)

    # CSV files
    if (file_fmt == ".csv") & (obj_class == "pandas"):
        with s3client.get_object(Bucket=bucket, Key=s3key) as s3_obj:
            obj = read_csv(BytesIO(s3_obj["Body"].read()), **kwargs)

    # Excel files
    elif (file_fmt == ".xlsx") & (obj_class == "pandas"):
        with s3client.get_object(Bucket=bucket, Key=s3key) as s3_obj:
            obj = read_excel(BytesIO(s3_obj["Body"].read()), **kwargs)

    # Pickle files
    elif (file_fmt == ".pkl") & (obj_class == "pandas"):
        with s3client.get_object(Bucket=bucket, Key=s3key) as s3_obj:
            obj = read_pickle(BytesIO(s3_obj["Body"].read()), **kwargs)

    # elif (file_fmt == ".pkl") & (obj_class == "sklearn"):
    #     with TemporaryFile() as temp_file:
    #         s3client.download_fileobj(Bucket=bucket, Key=s3key, 
    #                                   Fileobj=temp_file)
    #         temp_file.seek(0)
    #         obj = joblib.load(temp_file)

    # HDF5 files
    elif (file_fmt == ".h5") & (obj_class == "pandas"):
        with NamedTemporaryFile() as temp_file:
            s3client.download_fileobj(Bucket=bucket, Key=s3key, 
                                      Fileobj=temp_file)
            temp_file.seek(0)
            obj = read_hdf(temp_file.name, **kwargs)

    # npy files
    elif (file_fmt == ".npy") & (obj_class == "numpy"):
        with NamedTemporaryFile() as temp_file:
            s3client.download_fileobj(Bucket=bucket, Key=s3key, 
                                      Fileobj=temp_file)
            temp_file.seek(0)
            obj = load(temp_file.name)

    # npz files- This exports each pandas column to a compressed numpy array
    # as a "key-value pair". See numpy documentation for more information
    elif (file_fmt == ".npz") & (obj_class == "numpy"):
        with NamedTemporaryFile() as temp_file:
            s3client.download_fileobj(Bucket=bucket, Key=s3key,
                                      Fileobj=temp_file)
            temp_file.seek(0)
            obj = load(temp_file.name)

    # If none of the above file_fmt and obj_class combination are true. Raise!
    else:
        raise_str = f"""Failed! {file_fmt} export of {obj_class} objects not
            implemented."""
        logger.critical('\tFailed! %s', raise_str)
        raise Exception(raise_str)

    logger.info('\tOK! %s loaded in %s [s]', obj.shape[0], (time() - start))
    return obj


def list_files_in_path(bucket, prefix, drop_sub_folders=True, 
                       remove_prefix=True, subset_ext=None, remove_ext=False,):
    """List all files (sorted) in the specified bucket and prefix.

    Arguments:
        - bucket [str]          -   S3 bucket where path i located,
                                    no trailing "/" like "stage-data-scientist"
        - prefix [str]          -   S3 prefix where files should be listed
        - drop_sub_folders [bool] -   Switch if files in subfolders should
                                    be dropped (default = True)
        - remove_prefix [bool]   -   Switch if strings in file_list should
                                    have file prefix removed (default=True)
        - subset_ext [str]           String ('.csv','.xlsx',...) to subset file
                                    extension with. If provided (default=None)
                                    file_list will only return files
                                    with the specified file extension.
        - remove_ext [bool]      -   Switch if strings in file_list should
                                    have file extension ('.csv','.xlsx',...)
                                    removed (default = False)
    Returns:
        - file_list [lst]     -   List of files in bucket-prefix.
    """
    # Initiate Logger
    logger = getLogger(__name__)
    start = time()

    # Remove filename from prefix
    if prefix != "":
        prefix = f"{dirname(prefix)}/"

    logger.info('List files in:\n\tbucket=%s\n\tprefix=%s',
                bucket, prefix)

    # List all s3keys
    file_list = [
        objectSummary.key
        for objectSummary in list(
            resource("s3").Bucket(bucket).objects.filter(Prefix=prefix)
        )[0:]
    ]

    file_list = [f for f in file_list if f != prefix]

    # Drop all files in subfolders
    if drop_sub_folders:
        file_list = [f for f in file_list if
                     (dirname(f) == dirname(prefix))]

    if remove_prefix:
        file_list = [f.replace(prefix, "") for f in file_list]

    # Subset only files ending with provided subset_ext
    if subset_ext:
        file_list = [f for f in file_list if
                     subset_ext in f]

    # Remove all file extensions
    if remove_ext:
        file_list = [splitext(f)[0] for f in file_list]

    logger.info('\tOK! Elapsed time %f [s]', time() - start)
    return sorted(file_list)



"Defines all functions within the rds (relational database service) module."

# Standard library imports
from logging import getLogger
from traceback import format_exc
from time import time
from configparser import ConfigParser
from re import match

# Related third party imports 
import psycopg2
from sqlalchemy import create_engine, Table, MetaData
from sqlalchemy.engine import reflection
from numpy import dtype
from pandas import DataFrame, read_sql_query, to_datetime, to_numeric, \
    StringDtype, BooleanDtype, Int64Dtype, Int32Dtype


def config(file_path: str = "/home/ec2-user/SageMaker/.config/database.ini",
           section: str = "maximo") -> dict:
    """Get database credentials from stored .ini file.
    
    The following config() function reads in the database.ini file and
    returns the connection parameters as a dictionary. This function will be
    imported in to the main python script.

    Arguments:
        file_path       - Path to the database.ini-file required to initiate
                          connection.
        section         - Section of database.ini-file where the
                          connection-parameters are stored.
    Returns:
        db_cred         - All connection parameters in the specified file_path
                          and section

    """
    parser = ConfigParser()      # Create a parser
    parser.read(file_path)       # Read config file
    db_cred = {}                 # Get section, default to postgresqs

    if parser.has_section(section):  # Checks to see (section) parser exists
        params = parser.items(section)
        for param in params:
            db_cred[param[0]] = param[1]
    else:
        raise Exception(f'Section {section} not found in the {file_path} file')
    return db_cred


def check_connection(db_cred: dict):
    """Check connection settings against DB and print version of postgreSQL.

    Arguments:
        db_cred (dict)  - Connection parameters from config file or manually
                          defined dictionary

    """
    connection = None

    # Initiate Logger
    logger = getLogger(__name__)
    logger.info('Checking DB Connection')

    try:
        # Setup connection
        connection = psycopg2.connect(**db_cred)
        cursor = connection.cursor()

        # Print PostgreSQL Connection properties
        dsn_para = connection.get_dsn_parameters()

        # Print PostgreSQL version
        cursor.execute("SELECT version();")           
        record = cursor.fetchone()

        # Construct connection string
        logger.info("DSN Parameters: %s", dsn_para)
        logger.info("\tConnection information: %s", record)
  
    except psycopg2.Error as error:
        logger.error("Error while connecting to PostgreSQL, %s", error)
    finally:
        
        # Closing database connection.
        if connection:
            cursor.close()
            connection.close()
            logger.info("\tConnection is closed")


def import_data(sql_query: str, db_cred: dict, df_meta: DataFrame = None,
                post_rename: dict = None, convert_dtypes: bool = True,
                query_name: str = None) -> DataFrame:
    """Load data from RDS instance.

    The function runs the provided sql_query against the database specified 
    in the db_cred dictionary. If df_meta-DataFrame is provided and 
    convert_dtypes = True (default), the function will also convert all columns
    to optimal Pandas dtype according to the provided meta data.  

    Arguments:
        sql_query (str)         - SQL-query to run
        db_cred (dict)          - Connection parameters from config file
        df_meta (DataFrame)     - DataFrame with schema information from
                                  import_meta_data function
        post_rename (dict)      - Dictionary for renaming dictionaries after 
                                  dtype conversion
        convert_dtypes (bool)   - Flag if data types should be converted using
                                  the data base schema for the used table
        query_name (str)        - Name of query to be executed (only for 
                                  logging purposes).
    Returns:
        df_ (DataFrame)         - DataFrame with loaded data

    """
    connection = None

    # Initiate Logger
    logger = getLogger(__name__)
    start = time()

    if query_name:
        logger.info('Loading %s Data:', query_name)
    else:
        logger.info('Loading Data:')

    try:
        # Get Data
        connection = psycopg2.connect(**db_cred)
        cursor = connection.cursor()
        df_ = read_sql_query(sql_query, connection)

        if (df_meta is not None) & (convert_dtypes):
            for col, data_dtype in df_.dtypes.iteritems():

                # Check schema
                df_meta_sub = df_meta.loc[df_meta.name == col, ]

                if df_meta_sub.empty is False:

                    # If there are more than one schema type present in df_meta
                    if df_meta_sub.type.nunique() > 1:
                        schema_dtype = df_meta_sub.type.value_counts().index[0]
                        logger.warning("\t Warning:'%s' has multiple dtypes "
                                       "in schema. Using most popular dtype "
                                       " %s for conversion", col, schema_dtype)
                    else:
                        schema_dtype = df_meta_sub.type.unique()[0]

                    # Int64s
                    if schema_dtype == "BIGINT":
                        df_[col] = df_[col].astype(Int64Dtype())
                    
                    # Int32s
                    elif schema_dtype == "INTEGER":
                        df_[col] = df_[col].astype(Int32Dtype())

                    # Bools
                    elif schema_dtype == "BOOLEAN":
                        df_[col] = df_[col].astype(BooleanDtype())

                    # Float64s
                    elif schema_dtype == "DOUBLE PRECISION":
                        df_[col] = to_numeric(df_[col],
                                              errors="coerce",
                                              downcast="float")
                    # Float64s
                    elif bool(match(r"NUMERIC\([0-9]{1,6},\s[0-9]{1,6}\)",
                                    schema_dtype)):
                        df_[col] = to_numeric(df_[col],
                                              errors="coerce",
                                              downcast="float")

                    # PostgreSQL provides users with the interval data type 
                    # that allows users to store and manipulate a time period. 
                    elif schema_dtype == 'INTERVAL':
                        # Implement better solution for this dtype when needed
                        df_[col] = df_[col].astype(StringDtype()) 

                    # Dates and timestamps
                    elif schema_dtype == "DATE":
                        df_[col] = to_datetime(df_[col],
                                               infer_datetime_format=True,
                                               utc=True,
                                               errors="coerce")

                    elif schema_dtype == "TIME WITHOUT TIME ZONE":
                        if data_dtype == dtype('O'):
                            df_[col] = to_datetime(df_[col],
                                                   infer_datetime_format=True,
                                                   utc=True,
                                                   errors="coerce")

                    elif schema_dtype == "TIMESTAMP WITH TIME ZONE":
                        if data_dtype == dtype('O'):
                            df_[col] = to_datetime(df_[col],
                                                   infer_datetime_format=True,
                                                   utc=True,
                                                   errors="coerce")
                    # Strings 
                    elif (schema_dtype in ["TEXT"] 
                          or bool(match(r"VARCHAR\([0-9]{1,6}\)",
                                        schema_dtype))):
                        df_[col] = df_[col].astype(StringDtype())

                    # Raise error if conversion hasn't been implemented
                    else:
                        raise_str = f"""\tData conversion '{schema_dtype}' not
                                        implemented. Please contact package 
                                        administrator"""
                        logger.critical(raise_str)
                        raise Exception(raise_str)
                else:
                    logger.warning("\tNo dtype for %s found in schema", col)
        elif df_meta is None:
            logger.warning("""\tNo dtype conversion was executed. Meta data 
                              (df_meta) not passed to function""")

        # Remove Timezone info from datetime tags
        for col in df_.select_dtypes(include='datetime64[ns, UTC]').columns:
            df_[col] = df_[col].dt.tz_localize(None)

        # Post rename
        if post_rename:
            df_ = df_.rename(columns=post_rename)

        # Output results to logger and return string
        logger.info("\tOK! Query executed in %f [s]. Shape = %s""",
                    time() - start,
                    df_.shape)

    except Exception:
        traceback_str = format_exc()
        logger.critical("Failed to execute query (%s)", traceback_str)
        raise Exception

    finally:
        if connection:
            cursor.close()
            connection.close()
            logger.info("\tConnection is closed")

    return df_


def import_meta_data(db_cred: dict, 
                     like_str: str = "maximoworker") -> DataFrame:
    """Import schema to pandas DataFrame.
    
    The functions main purpose is to import data type for all tables containing
    the like_str argument in the db accociated with the passed credentials. 

    Arguments:
        db_cred (dict)  - Dictionary containing all login information either 
                          manually entered or using the config function of 
                          the exergi-package.
        like_str (str)  - List all tables with table_name starting with 
                          this string. (default: "maximoworker")

    Keyword Arguments:
        kwargs          - Key word arguments passed to create_engine from 
                          sqlalchemy

    Returns:
        df_meta   - DataFrame containing

    """
    # Initiate Logger
    logger = getLogger(__name__)
    logger.info('Loading Meta Data:')
    start = time()

    try:
        engine = create_engine(
            'postgresql+psycopg2://{}:{}@{}/{}'.format(
                db_cred["user"],
                db_cred["password"],
                db_cred["host"],
                db_cred["database"]))

        # Get metadata
        metadata = MetaData(bind=engine)
        insp = reflection.Inspector.from_engine(engine)
        tables = [t for t in insp.get_table_names(schema='public')
                  if like_str in t]

        # Store metadata in Pandas DataFrame
        df_meta = DataFrame(
            data=[{"table": t, "name": c.name, "type": str(c.type),
                   "python_type": c.type.python_type,
                   "nullable": c.nullable} for t in tables for c in
                  Table(t, metadata, autoload=True,
                        autoload_with=engine).columns])
        
        # Set dtypes
        df_meta = df_meta.astype({"table": StringDtype(),
                                  "name": StringDtype(),
                                  "type": StringDtype(), 
                                  "nullable": bool})
        # Log results
        logger.info("""\tOK! Query executed in %.3f [s].
                    Shape = %s""", (time() - start), df_meta.shape)

    except Exception as error:
        logger.info("\tError while querying data, %s", error)
        raise error

    finally:
        if engine:
            logger.info("\tConnection is closed")
            engine.dispose()

    return df_meta



"""
Function for building the exergi package.

To build package use:
    python setup.py sdist bdist_wheel
To upload package to PyPi:
    twine upload dist/*
"""

import pathlib
from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent  # The directory containing this file
README = (HERE / "README.md").read_text()  # The text of the README file

links = []
requires = []

# This call to setup() does all the work
setup(
    name="exergi",
    version="0.4.28",
    description="SE Libary for various tasks",
    long_description=README,
    long_description_content_type="text/markdown",
    author="KasperJanehag",
    author_email="kasper.janehag@gmail.com",
    license="MIT",
    classifiers=["License :: OSI Approved :: MIT License",
                 "Programming Language :: Python :: 3",
                 "Programming Language :: Python :: 3.7"],
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=["pandas>=1.1.1",
                      "boto3>=1.14.15",
                      "tables",
                      "numpy>=1.19.0",
                      "psycopg2-binary",
                      "sqlalchemy",
                      "plotly"],
)
